import importlib
import os
from typing import Optional

from config import get_config
from fastapi import APIRouter
from fastapi.routing import Mount
from fastapi.staticfiles import StaticFiles
from libs.api.api_handler import ApiHandler

config = get_config()


class Bootstrap:
    def __init__(self) -> None:
        self.controllers = []
        self.current_path: str = os.path.join(
            os.path.relpath(os.path.join(os.getcwd())), '${project_name}'
        )
        path = self.get_project_path()
        modules_path = self.get_module_path(path)
        for root, _, files in os.walk(modules_path):
            for file in files:
                if 'controller' in file.lower() and file.endswith('.py'):
                    module_path = (
                        root.replace('/', '.')[2:]
                        + '.'
                        + file.replace('.py', '')
                    )
                    module_path = '.'.join(module_path.split('.')[1:])
                    modules = importlib.import_module(module_path)

                    for d in dir(modules):
                        router = getattr(modules, d)
                        if isinstance(router, APIRouter):
                            self.controllers.append(router)

    def get_project_path(self) -> Optional[str]:
        for root, dirs, _ in os.walk(self.current_path):
            if root.startswith('./.') or root.endswith('.'):
                continue
            if 'modules' in dirs:
                return root

    def get_module_path(self, path: str) -> Optional[str]:
        for root, dirs, files in os.walk(path):
            if (
                '__pycache__' in dirs
                or '__pycache__' in root
                or 'modules' not in root
            ):
                continue
            return root

    def setup_fastapi_app(self, app: ApiHandler):
        for controller in self.controllers:
            app.add_router(controller)



import os
import pathlib
import shutil

from mako.template import Template

from ...config import DatabaseConfig, ModuleConfig, UserConfig


def run(
    project_name: str,
    module_config: ModuleConfig,
    user_config: UserConfig,
    database_config: DatabaseConfig,
):
    current_dir = os.path.dirname(__file__)
    shutil.copytree(
        os.path.join(current_dir, 'libs/'), os.path.join(project_name, 'libs')
    )
    template = Template(
        open(
            os.path.join(current_dir, 'libs', 'bootstrap', 'bootstrap.py')
        ).read()
    )
    boostrap_content = template.render(project_name=project_name)
    with open(
        os.path.join(project_name, 'libs', 'bootstrap', 'bootstrap.py'), 'w'
    ) as file:
        file.write(boostrap_content)
    shutil.copytree(
        os.path.join(current_dir, 'utils/'),
        os.path.join(project_name, 'utils'),
    )
    os.makedirs(os.path.join(project_name, module_config.module_path))
    shutil.copytree(
        os.path.join(current_dir, 'modules/user/'),
        os.path.join(project_name, module_config.module_path, 'user'),
    )
    shutil.copy(
        os.path.join(current_dir, 'config.py'),
        os.path.join(project_name, 'config.py'),
    )
    shutil.copy(
        os.path.join(current_dir, 'main.py'),
        os.path.join(project_name, 'main.py'),
    )
    shutil.copy(
        os.path.join(current_dir, 'app.py'),
        os.path.join(project_name, 'app.py'),
    )
    if not os.path.isfile(os.path.join(current_dir, '.env')):
        with open('.env', 'w') as file:
            file.write('')
    with open('.env', 'a') as file:
        template = Template(open(os.path.join(current_dir, '.env')).read())
        content = template.render(project_name=project_name)
        file.write(f'\n{content}')
    print(
        f'Please, install the packages: fastapi pyjwt argon2-cffi sqlalchemy python-dotenv'
    )


def new_module(module_name: str):
    ...



from datetime import datetime, timedelta
from functools import cache
from typing import Optional, TypedDict

import jwt
from argon2 import PasswordHasher
from argon2.exceptions import InvalidHashError, VerifyMismatchError
from config import get_config
from fastapi import HTTPException, Request, Response, status
from jwt.exceptions import ExpiredSignatureError
from libs.database.database_handler import DatabaseHandler
from modules.user.models import UserModel
from pydantic import ValidationError
from sqlalchemy import select


class AccessTokenDict(TypedDict):
    access_token: str
    refresh_access_token: str


class _Payload(TypedDict):
    exp: str
    sub: str


class Security:

    argon = PasswordHasher()
    config = get_config()
    database_handler = DatabaseHandler()

    def __init__(self) -> None:
        self.JWT_SECRET_KEY = self.config.JWT_SECRET_KEY
        self.JWT_REFRESH_SECRET_KEY = self.config.JWT_REFRESH_SECRET_KEY
        self.JWT_ALGORITHM = self.config.JWT_ALGORITHM
        self.ACCESS_TOKEN_EXPIRE_MINUTES = (
            self.config.ACCESS_TOKEN_EXPIRE_MINUTES
        )
        self.REFRESH_TOKEN_EXPIRE_MINUTES = (
            self.config.REFRESH_TOKEN_EXPIRE_MINUTES
        )

    def crypt_password(self, password: str) -> str:
        encrypted_password = self.argon.hash(password)
        return encrypted_password

    def verify_password(self, hash: str, password: str) -> bool:
        return self._verify_password(hash, password)

    @cache
    def _verify_password(self, hash: str, password: str) -> bool:
        try:
            self.argon.verify(hash, password)
        except (VerifyMismatchError, InvalidHashError):
            return False
        return True

    def _generate_jwt(self, subject: str, expires: datetime) -> str:
        to_encode = {'exp': expires, 'sub': str(subject)}
        enconded_jwt = jwt.encode(
            to_encode, self.JWT_SECRET_KEY, self.JWT_ALGORITHM
        )
        return enconded_jwt

    def create_access_token(self, subject: str) -> str:
        expires_delta = datetime.utcnow() + timedelta(
            minutes=self.ACCESS_TOKEN_EXPIRE_MINUTES
        )
        return self._generate_jwt(subject, expires_delta)

    def create_refresh_access_token(self, subject: str) -> str:
        expires_delta = datetime.utcnow() + timedelta(
            minutes=self.REFRESH_TOKEN_EXPIRE_MINUTES
        )
        return self._generate_jwt(subject, expires_delta)

    def decode_jwt(self, token: str) -> _Payload:
        return jwt.decode(
            token, self.JWT_SECRET_KEY, algorithms=[self.JWT_ALGORITHM]
        )

    def get_current_user(
        self, request: Request, response: Response
    ) -> UserModel:
        if not (tokens := self._get_access_token(request)):
            raise HTTPException(
                status.HTTP_400_BAD_REQUEST, 'token not provided'
            )
        try:
            payload = self.decode_jwt(tokens['access_token'])
        except (jwt.PyJWKError, ValidationError, jwt.DecodeError):
            raise HTTPException(
                status_code=status.HTTP_403_FORBIDDEN,
                detail='Could not validate credentials',
                headers={'WWW-Authenticate': 'Bearer'},
            )
        except ExpiredSignatureError:
            try:
                payload = self.decode_jwt(tokens['refresh_access_token'])
                access_token = self._renew_access_token_from_refresh(payload)
                response.set_cookie('access_token', access_token)
            except ExpiredSignatureError:
                raise HTTPException(
                    status_code=status.HTTP_403_FORBIDDEN,
                    detail='JWT signature has expired',
                    headers={'WWW-Authenticate': 'Bearer'},
                )
        if current_user := self._get_user_model_by_uuid_subject(
            payload['sub']
        ):
            if not current_user:
                raise HTTPException(
                    status.HTTP_403_FORBIDDEN, 'User not found'
                )

        return current_user

    def _get_user_model_by_uuid_subject(
        self, uuid: str
    ) -> Optional[UserModel]:
        with self.database_handler as session:
            query = select(UserModel).where(UserModel.uuid == uuid)
            current_user = session.scalar(query)
        return current_user

    def _get_access_token(self, request: Request) -> Optional[AccessTokenDict]:
        if 'access_token' in request.cookies.keys():
            return {
                'access_token': request.cookies.get('access_token'),
                'refresh_access_token': request.cookies.get(
                    'refresh_access_token'
                ),
            }
        elif 'access_token' in request.headers.keys():
            return {
                'access_token': request.headers.get('access_token'),
                'refresh_access_token': request.headers.get(
                    'refresh_access_token'
                ),
            }

    def _renew_access_token_from_refresh(self, refresh_token: _Payload) -> str:
        return self.create_access_token(refresh_token['sub'])



import uuid
from datetime import datetime

from libs.database import Base
from sqlalchemy import String
from sqlalchemy.orm import Mapped, mapped_column


class TodoModel(Base):

    __tablename__ = 'todos'

    id: Mapped[int] = mapped_column(init=False, primary_key=True)
    uuid: Mapped[str] = mapped_column(
        String(60), init=False, default_factory=lambda: str(uuid.uuid4())
    )

    name: Mapped[str] = mapped_column(String(60), nullable=False)



from libs.security import Security
from sqlalchemy import event
from sqlalchemy.engine.base import Connection
from sqlalchemy.orm.mapper import Mapper

from .models import UserModel


class UserEvents:
    @staticmethod
    @event.listens_for(UserModel, 'before_insert', propagate=True)
    def my_listener_function(
        mapper: Mapper, connection: Connection, target: UserModel
    ):
        target.password = Security().crypt_password(target.password)



# Control
class UserNotFound(Exception):
    ...


class UserAlreadyExists(Exception):
    ...


# Invalid
class InvalidUsername(Exception):
    ...


class InvalidEmail(Exception):
    ...


class InvalidGender(Exception):
    ...


class InvalidPassword(Exception):
    ...


class InvalidName(Exception):
    ...


# Login
class InvalidUserLogin(Exception):
    ...



import uvicorn
from config import get_config

if __name__ == '__main__':
    config = get_config()
    uvicorn.run(
        'app:app',
        host=config.API_HOST,
        port=config.API_PORT,
        reload=True,
    )



from libs.database import DatabaseHandler
from libs.security import Security

from .ext import (
    InvalidPassword,
    InvalidUserLogin,
    UserAlreadyExists,
    UserNotFound,
)
from .models import UserModel
from .schemas import LoginUserRequest, RegisterUserRequest
from .service import UserService


class UserRepository:

    database_handler = DatabaseHandler()
    service = UserService()
    secutiry = Security()

    def login_user(self, data: LoginUserRequest) -> UserModel:
        if len(data.user) < 4:
            raise InvalidUserLogin('Login user not provided')
        self.service.validate_password(data.password)

        user = self.service.get_user_by_email_or_username(data.user, data.user)
        if not user:
            raise UserNotFound('User not found')
        if not self.secutiry.verify_password(user.password, data.password):
            raise InvalidPassword('Password Wrong')

        return user

    def register_user(self, data: RegisterUserRequest) -> UserModel:
        self.service.validate_register_data(data)
        if any(self.service.valid_email_or_username_in_use(data).values()):
            raise UserAlreadyExists('User already exists')
        with self.database_handler as session:
            gender = self.service.get_gender_by_name(data.gender)
            new_user = UserModel(gender_id=gender.id, **data.model_dump())
            session.add(new_user)
            session.commit()
            session.refresh(new_user)
            return new_user
